
#ifndef AXIS_TYPES_H
#define AXIS_TYPES_H


#include <memory> // for |std::unique_ptr|, |std::shared_ptr|
#include "../Atlas.h" 

#include "sl_list.h" 

#include <iostream> 

#include <vector>
#include <cassert>

#include <iterator>

#include <stdexcept>

namespace atlas { namespace interpreter {

struct type_expr;
typedef type_expr* type_p;
typedef const type_expr* const_type_p;
typedef std::unique_ptr<type_expr> type_ptr;

typedef containers::simple_list<type_expr> type_list;
typedef atlas::containers::sl_node<type_expr>* raw_type_list;
typedef containers::sl_list<type_expr> dressed_type_list;
typedef containers::weak_sl_list_const_iterator<type_expr> wtl_const_iterator;
typedef containers::weak_sl_list_iterator<type_expr> wtl_iterator;
  

enum type_tag 
{ undetermined_type, primitive_type, row_type, tuple_type, function_type };

enum primitive_tag
{ integral_type, rational_type, string_type, boolean_type
  , 
    vector_type, matrix_type, rational_vector_type,
    complex_lie_type_type , root_datum_type, inner_class_type, real_form_type,
    Cartan_class_type, KGB_element_type, block_type,
    module_parameter_type, split_integer_type, virtual_module_type, 
 
  nr_of_primitive_types };

struct func_type; // must be predeclared for |type_expr|

struct type_expr
{ type_tag kind;
  union
  { primitive_tag prim; // when |kind==primitive|
    type_p component_type; // when |kind==row_type|
    raw_type_list tupple; // when |kind==tuple_type|
    func_type* func; // when |kind==function_type|
  };

  
  type_expr() noexcept : kind(undetermined_type) {}
  explicit type_expr(primitive_tag p) noexcept
    : kind(primitive_type), prim(p) {}
  explicit type_expr(type_ptr&& c) noexcept
    : kind(row_type), component_type(c.release()) {}
  explicit type_expr(type_list&& l) noexcept; // tuple types
  explicit type_expr(dressed_type_list&& l) noexcept; // tuple types
  inline type_expr(type_expr&& arg, type_expr&& result) noexcept;
   // for function types
  
  #ifdef incompletecpp11
  type_expr(const type_expr& t) = delete;
  type_expr& operator=(const type_expr& t) = delete;
  #endif
  type_expr(type_expr&& t) noexcept; // move constructor
  type_expr& operator=(type_expr&& t) noexcept; // do move assignment only
  void clear() noexcept;
    // resets to undefined state, cleaning up owned pointers
  ~type_expr() noexcept { clear(); }
    // that is all explcitly needed for destruction
  void swap(type_expr& t) noexcept;
  
  type_expr copy() const; // in lieu of deep copy constructor
  void set_from(type_expr&& p) noexcept; // shallow copy
  bool specialise(const type_expr& pattern);
    // try to match pattern, possibly modifying |*this|
  bool can_specialise(const type_expr& pattern) const;
    

 };


struct func_type
{ type_expr arg_type, result_type;

  func_type(type_expr&& a, type_expr&& r)
 : arg_type(std::move(a)), result_type(std::move(r)) {}
#ifdef incompletecpp11
  func_type(const func_type& f) = delete;
  func_type& operator=(const func_type& f) = delete;
  func_type(func_type&& f)
  : arg_type(std::move(f.arg_type)), result_type(std::move(f.result_type))
  {}
  func_type& operator=(func_type&& f)
  { arg_type = std::move(f.arg_type); result_type = std::move(f.result_type);
    return *this;
  }
#else
  func_type(func_type&& f) = default; // move constructor
  func_type& operator=(func_type&& f) = default; // move assignment
#endif
  func_type copy() const // in lieu of a copy contructor
  { return func_type(arg_type.copy(),result_type.copy()); }
};
typedef func_type* func_type_p;
typedef std::unique_ptr<func_type> func_type_ptr;

struct value_base
{ value_base()  {};
  virtual ~value_base() = 0;
  virtual void print(std::ostream& out) const =0;
  virtual value_base* clone() const { assert(false); return nullptr; }
  static const char* name(); // just a model; this instance remains undefined
protected:
#ifndef incompletecpp11
  value_base(const value_base& x) = default;
#else
  value_base(const value_base& x) {}
#endif
public:
  value_base& operator=(const value_base& x) = delete;
};
inline value_base::~value_base() {} // necessary but empty implementation

typedef const value_base* value;
typedef std::shared_ptr<const value_base> shared_value;
typedef std::shared_ptr<value_base> own_value;

struct row_value : public value_base
{ std::vector<shared_value> val;

  explicit row_value(size_t n) : val(n) {} // start with |n| null pointers
  template <typename I> row_value(I begin, I end) : val(begin,end)
    {} // set from iterator range
  void print(std::ostream& out) const;
  size_t length() const { return val.size(); }
  row_value* clone() const { return new row_value(*this); }
    // copy the outer level vector
  static const char* name() { return "row value"; }
protected:
  row_value(const row_value& v) : val(v.val) {}
    // copy still shares the individual entries
};

typedef std::shared_ptr<const row_value> shared_row;
typedef std::shared_ptr<row_value> own_row;

struct tuple_value : public row_value
{ tuple_value(size_t n) : row_value(n) {}
  template <typename I> tuple_value(I begin, I end) : row_value(begin,end) {}
  tuple_value* clone() const { return new tuple_value(*this); }
  void print(std::ostream& out) const;
  static const char* name() { return "tuple value"; }
private:
  tuple_value(const tuple_value& v) : row_value(v){}
 // copy constructor; used by |clone|
};

typedef std::unique_ptr<tuple_value> tuple_ptr;
typedef std::shared_ptr<const tuple_value> shared_tuple;
typedef std::shared_ptr<tuple_value> own_tuple;

typedef std::shared_ptr<class evaluation_context> shared_context;
class evaluation_context
{ shared_context next;
  std::vector<shared_value> frame;
  evaluation_context(const evaluation_context&) = delete;
  // never copy contexts
public:
  evaluation_context (const shared_context& next)
: next(next), frame() {}
  void reserve (size_t n) { frame.reserve(n); }
  shared_value& elem(size_t i,size_t j);
  std::back_insert_iterator<std::vector<shared_value> > back_inserter ()
  { return std::back_inserter(frame); }
  const shared_context& tail() const { return next; }
  std::vector<shared_value>::const_iterator begin() const
    { return frame.begin(); }
  std::vector<shared_value>::const_iterator end() const
    { return frame.end(); }
};

struct expression_base
{ enum level { no_value, single_value, multi_value };

  expression_base()  {}
  expression_base(const expression_base&) = delete; // they are never copied
  expression_base(expression_base&&) = delete; // nor moved
  expression_base& operator=(const expression_base&)=delete; // nor assigned
  expression_base& operator=(expression_base&&)=delete; // nor move-assigned
  virtual ~expression_base()  {}
// other virtual methods
  virtual void evaluate(level l) const =0;
  virtual void print(std::ostream& out) const =0;
// non-virtuals that call the virtual |evaluate|
  void void_eval() const { evaluate(no_value); }
  void eval() const { evaluate(single_value); }
  void multi_eval() const { evaluate(multi_value); }
};

typedef expression_base* expression;
typedef std::unique_ptr<expression_base> expression_ptr;
typedef std::shared_ptr<expression_base> shared_expression;

struct conversion_info
{ typedef void (*conv_f)();
  conv_f convert;
  std::string name;
  conversion_info(const char* s,conv_f c) : convert(c),name(s) {}
};

class conversion : public expression_base
{ const conversion_info& type;
  expression_ptr exp;
public:
  conversion(const conversion_info& t,expression_ptr e)
   :type(t),exp(std::move(e)) {}
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct conversion_record : public conversion_info
{ const type_expr* from, * to;
  // non-owned pointers, themselves could be |const| in gcc 4.8
  conversion_record  (const type_expr& from_type,
                     const type_expr& to_type,
                     const char* s, conv_f c)
   : conversion_info(s,c), from(&from_type),to(&to_type) {}
};

class voiding : public expression_base
{ expression_ptr exp;
public:
  voiding(expression_ptr e) : exp(e.release()) {}
  virtual void evaluate(level l) const;
  virtual void print(std::ostream& out) const;
};

struct error_base : public std::exception
{ std::string message;
  explicit error_base(const std::string& s) : message(s) {}
#ifdef incompletecpp11
  ~error_base () throw() {} // backward compatibility for gcc 4.4
#endif
  const char* what() const throw() { return message.c_str(); }
};

struct logic_error : public error_base
{ explicit logic_error(const std::string& s) : error_base(s) {}
#ifdef incompletecpp11
  ~logic_error () throw() {} // backward compatibility for gcc 4.4
#endif
};

struct program_error : public error_base
{ explicit program_error(const std::string& s) : error_base(s) {}
#ifdef incompletecpp11
  ~program_error () throw() {} // backward compatibility for gcc 4.4
#endif
};

struct runtime_error : public error_base
{ explicit runtime_error(const std::string& s) : error_base(s) {}
#ifdef incompletecpp11
  ~runtime_error () throw() {} // backward compatibility for gcc 4.4
#endif
};

struct expr; // predeclare
struct expr_error : public program_error
{ const expr& offender; // the subexpression causing a problem

  expr_error (const expr& e,const std::string& s) throw()
    : program_error(s),offender(e) {}
#ifdef incompletecpp11
  ~expr_error() throw() {}
#endif
};

struct type_error : public expr_error
{ type_expr actual; 
  type_expr required; // the types that conflicted

  type_error (const expr& e, type_expr&& a, type_expr&& r) noexcept 
    : expr_error(e,"Type error") 
      ,actual(std::move(a)),required(std::move(r)) {}
#ifdef incompletecpp11
  type_error(type_error&& e)
  : expr_error(std::move(e))
  , actual(std::move(e.actual)), required(std::move(e.required)) {}
  ~type_error () throw() {}
#else
  type_error(type_error&& e) = default;
#endif
};

struct balance_error : public expr_error
{ containers::sl_list<type_expr> variants;
  balance_error(const expr& e)
  : expr_error(e,"No common type found"), variants() {}
#ifdef incompletecpp11
  balance_error(const balance_error&)=delete;
  balance_error(balance_error&& o) 
  : expr_error(std::move(o)),variants(std::move(o.variants)) {}
  ~balance_error() throw() {}
#endif
};

struct loop_break : private logic_error
{ unsigned depth;
  loop_break(int n)
  : logic_error("Uncaught break from loop"), depth(n) {}
#ifdef incompletecpp11
  ~loop_break () throw() {}
#endif
} ;

struct function_return : private logic_error
{ shared_value val;
  function_return(shared_value&& val)
  : logic_error("Uncaught return from function"), val(val) {}
#ifdef incompletecpp11
  ~function_return () throw() {}
#endif
} ;


extern const char* prim_names[];

extern const type_expr unknown_type; // \.{*}
extern const type_expr void_type; // \.{()}
extern const type_expr int_type; // \.{int}
extern const type_expr bool_type; // \.{bool}
extern const type_expr row_of_type; // \.{[*]}
extern const type_expr gen_func_type; 

extern const type_expr rat_type; // \.{rat}
extern const type_expr str_type; // \.{string}
extern const type_expr vec_type; // \.{vec}
extern const type_expr ratvec_type; // \.{ratvec}
extern const type_expr mat_type; // \.{mat}
extern const type_expr row_of_int_type; // \.{[int]}
extern const type_expr row_of_rat_type; // \.{[rat]}
extern const type_expr row_of_vec_type; // \.{[vec]}
extern const type_expr row_row_of_int_type; // \.{[[int]]}
extern const type_expr pair_type; // \.{(*,*)}
extern const type_expr int_int_type; // \.{(int,int)}
extern const type_expr Lie_type_type; // \.{LieType}
extern const type_expr rd_type; // \.{RootDatum}
extern const type_expr ic_type; // \.{InnerClass}
extern const type_expr rf_type; // \.{RealForm}
extern const type_expr split_type; // \.{Split}
extern const type_expr param_type; // \.{RealForm}
extern const type_expr param_pol_type; 

extern std::vector<shared_value> execution_stack;


type_ptr acquire(const type_expr* t);

type_list empty_tuple();
type_list& prefix(type_expr&& t, type_list& dst);
dressed_type_list& prefix(type_expr&& t, dressed_type_list& dst);

std::ostream& operator<<(std::ostream& out, const type_expr& t);
std::ostream& operator<<(std::ostream& out, const func_type& f);

bool operator== (const type_expr& x,const type_expr& y);
inline bool operator!= (const type_expr& x,const type_expr& y)
{ return !(x==y); }

type_ptr mk_prim_type(primitive_tag p);
type_ptr mk_row_type(type_ptr&& c);
type_ptr mk_tuple_type (type_list&& l);
type_ptr mk_function_type(type_expr&& a, type_expr&& r);

type_p make_prim_type(unsigned int p);
type_p make_row_type(type_p c);
type_p make_tuple_type(raw_type_list l);
type_p make_function_type(type_p a,type_p r);

raw_type_list make_type_singleton(type_p raw);
raw_type_list make_type_list(raw_type_list l,type_p t);

type_ptr mk_type(const char* s);
type_expr mk_type_expr(const char* s);
  

type_expr unknown_tuple(size_t n);

std::ostream& operator<< (std::ostream& out, const value_base& v);

void push_tuple_components();
void wrap_tuple(size_t n);

inline std::ostream& operator<< (std::ostream& out, const expression_base& e)
{ e.print(out); return out; }

void push_expanded(expression_base::level l, const shared_value& v);

bool coerce(const type_expr& from_type, const type_expr& to_type,
            expression_ptr& e);
expression_ptr conform_types
  (const type_expr& found, type_expr& required
  , expression_ptr&& d, const expr& e);
const conversion_record* row_coercion(const type_expr& final_type,
				     type_expr& component_type);
void coercion(const type_expr& from,
              const type_expr& to,
              const char* s, conversion_info::conv_f f);

unsigned int is_close (const type_expr& x, const type_expr& y);

bool broader_eq (const type_expr& a, const type_expr& b);


type_expr::type_expr(type_expr&& arg, type_expr&& result) noexcept
 : kind(function_type)
   , func(new func_type(std::move(arg),std::move(result)))
   {}

template<typename T> T& as_lvalue(T&& rvalue) { return rvalue; }

inline void push_value(const shared_value& v)
  { execution_stack.push_back(v); }

inline void push_value(shared_value&& v)
  { execution_stack.push_back(std::move(v)); }

inline shared_value pop_value()
{ shared_value arg(std::move(execution_stack.back()));
  execution_stack.pop_back();
  return arg;
}

template <typename D> // |D| is a type derived from |value_base|
 inline std::shared_ptr<const D> get() throw(logic_error)
{ std::shared_ptr<const D> p=std::dynamic_pointer_cast<const D>(pop_value());
  if (p.get()==nullptr)
    throw logic_error(std::string("Argument is no ")+D::name());
  return p;
}



template <typename D> // |D| is a type derived from |value_base|
  inline std::shared_ptr<D> non_const_get() throw(logic_error)
{ return std::const_pointer_cast<D>(get<D>()); }

template <typename D> // |D| is a type derived from |value_base|
  D* force (value_base* v) throw(logic_error)
{ D* p=dynamic_cast<D*>(v);
  if (p==nullptr) throw
    logic_error(std::string("forced value is no ")+D::name());
  return p;
}

template <typename D> // |D| is a type derived from |value_base|
  const D* force (value v) throw(logic_error)
{ const D* p=dynamic_cast<const D*>(v);
  if (p==nullptr) throw
    logic_error(std::string("forced value is no ")+D::name());
  return p;
}

inline value_base* uniquify(shared_value& v)
{ if (not v.unique())
     v=shared_value(v->clone());
  return const_cast<value_base*>(v.get());
}

template <typename D> // |D| is a type derived from |value_base|
  std::shared_ptr<D> get_own()
{ std::shared_ptr<const D> p=get<D>();
  if (p.unique())
    return std::const_pointer_cast<D>(p);
  return std::shared_ptr<D>(p->clone());
}

template<unsigned int> void wrap_tuple();
template<unsigned int>
  void do_wrap(std::vector<shared_value>::iterator it);
template<>
   inline void do_wrap<0u>(std::vector<shared_value>::iterator it) {}
template<unsigned int n>
   inline void do_wrap(std::vector<shared_value>::iterator it)
   { *--it = pop_value(); do_wrap<n-1>(it); }
template<unsigned int n>
   inline void wrap_tuple()
   { std::shared_ptr<tuple_value> result = std::make_shared<tuple_value>(n);
     do_wrap<n>(result->val.end());
     push_value(std::move(result));
   }

} }
#endif

